"""Utilities for watsonx SQL integration."""

import contextlib
import urllib.parse
from collections.abc import Iterable
from typing import Any

try:
    from pyarrow import flight  # type: ignore[import-untyped]
except ModuleNotFoundError as e:
    error_msg = (
        "To use WatsonxSQLDatabase one need to install langchain-ibm with extras "
        "`sql_toolkit`: `pip install langchain-ibm[sql_toolkit]`",
    )
    raise ModuleNotFoundError(error_msg) from e

from ibm_watsonx_ai import APIClient, Credentials  # type: ignore[import-untyped]
from ibm_watsonx_ai.helpers.connections.flight_sql_service import (  # type: ignore[import-untyped]
    FlightSQLClient,
)
from langchain_core.utils.utils import from_env


def truncate_word(content: Any, *, length: int, suffix: str = "...") -> str | Any:
    """Truncate a string to a maximum length.

    Based on the analogous function from langchain_common.utilities.sql_database.py.
    """
    if not isinstance(content, str) or length <= 0:
        return content

    if len(content) <= length:
        return content

    return content[: length - len(suffix)].rsplit(" ", 1)[0] + suffix


def _validate_param(value: str | None, key: str, env_key: str) -> None:
    if value is None:
        error_msg = (
            f"Did not find {key}, please add an environment variable"
            f" `{env_key}` which contains it, or pass"
            f" `{key}` as a named parameter."
        )
        raise ValueError(error_msg)


def _from_env(env_var_name: str) -> str | None:
    """Read env variable. If it is not set, return None."""
    return from_env(env_var_name, default=None)()


def pretty_print_table_info(
    schema: str, table_name: str, table_info: dict[str, Any]
) -> str:
    """Pretty print table info."""

    def convert_column_data(field_metadata: dict[str, Any]) -> str:
        name = field_metadata.get("name")

        field_metadata_type = field_metadata.get("type", {})
        native_type = field_metadata_type.get("native_type")
        nullable = field_metadata_type.get("nullable")

        return f'"{name}" {native_type}{"" if nullable else " NOT NULL"}'

    create_table_template = """
CREATE TABLE "{schema}"."{table_name}" (
\t{column_definitions}{primary_key}{foreign_key}
\t)"""

    extended_metadata = table_info.get("extended_metadata", [{}])

    def _retrieve_field_data(field_name: str) -> dict[str, Any]:
        return next(
            filter(
                lambda el: el.get("name") == field_name,
                extended_metadata,
            ),
            {},
        )

    # Primary Key
    primary_key: dict[str, Any] = _retrieve_field_data("primary_key")
    if primary_key:
        key_columns = ", ".join(primary_key.get("value", {}).get("key_columns", []))
        primary_key_text = (
            f",\n\tCONSTRAINT {primary_key['name']} PRIMARY KEY ({key_columns})"
        )
    else:
        primary_key_text = ""

    # Foreign keys
    foreign_keys: dict[str, Any] = _retrieve_field_data("foreign_keys")
    if foreign_keys:
        foreign_keys_text = ""
        foreign_key_text_template = (
            "CONSTRAINT {fk_name} FOREIGN KEY ({col_name}) "
            "REFERENCES {external_table_name}({external_col_name})"
        )
        for foreign_key in foreign_keys.get("value", []):
            foreign_keys_text += ",\n\t"
            join_condition = foreign_key["join_condition"].split("=")
            foreign_keys_text += foreign_key_text_template.format(
                fk_name=foreign_key["name"],
                col_name=join_condition[0].strip().split(".")[-1],
                external_table_name=join_condition[1].strip().split(".")[1],
                external_col_name=join_condition[1].strip().split(".")[2],
            )
    else:
        foreign_keys_text = ""

    return create_table_template.format(
        schema=schema,
        table_name=table_name,
        column_definitions="\n\t".join(
            [
                # Do not add comma for the last column
                convert_column_data(field_metadata=field_metadata) + ","
                if index < len(table_info["fields"])
                else convert_column_data(field_metadata=field_metadata)
                for index, field_metadata in enumerate(table_info["fields"], start=1)
            ],
        ),
        primary_key=primary_key_text,
        foreign_key=foreign_keys_text,
    )


class WatsonxSQLDatabase:
    """Watsonx SQL Database class for IBM watsonx.ai databases connection assets.

    Uses Arrow Flight to interact with databases via watsonx.

    Args:
        connection_id: ID of db connection asset
        schema: name of the database schema from which tables will be read
        project_id: ID of project
        space_id: ID of space
        url: URL to the Watson Machine Learning or CPD instance
        api_key: API key to the Watson Machine Learning or CPD instance
        apikey: API key to the Watson Machine Learning or CPD instance (deprecated)
        token: service token, used in token authentication
        password: password to the CPD instance
        username: username to the CPD instance
        instance_id: instance_id of the CPD instance
        version: version of the CPD instance
        verify: certificate verification flag
        watsonx_client: instance of `ibm_watsonx_ai.APIClient`
        ignore_tables: list of tables that will be ignored
        include_tables: list of tables that should be included
        sample_rows_in_table_info: number of first rows to be added to the table info
        max_string_length: max length of string


    ???+ info "Setup"

        To use, you should have `langchain_ibm` python package installed,
        and the environment variable `WATSONX_API_KEY` set with your API key, or pass
        it as a named parameter `api_key` to the constructor.

        ```bash
        pip install -U langchain-ibm

        # or using uv
        uv add langchain-ibm
        ```

        ```bash
        export WATSONX_API_KEY="your-api-key"
        ```

        !!! deprecated
            `apikey` and `WATSONX_APIKEY` are deprecated and will be removed in
            version `2.0.0`. Use `api_key` and `WATSONX_API_KEY` instead.

    ??? info "Instantiate"

        ```python
        from langchain_ibm.utilities.sql_database import WatsonxSQLDatabase

        wx_sql_database = WatsonxSQLDatabase(
            connection_id="<CONNECTION_ID>",
            schema="<SCHEMA>",
            url="<URL>",
            project_id="<PROJECT_ID>",
            api_key="<API_KEY>",
        )
        ```

    ??? warning "Raises"
        - `ValueError` - if some required credentials are missing
        - `RuntimeError` - if no tables found in given schema

    """

    def __init__(
        self,
        *,
        connection_id: str,
        schema: str,
        project_id: str | None = None,
        space_id: str | None = None,
        url: str | None = None,
        api_key: str | None = None,
        apikey: str | None = None,
        token: str | None = None,
        password: str | None = None,
        username: str | None = None,
        instance_id: str | None = None,
        version: str | None = None,
        verify: str | bool | None = None,
        watsonx_client: APIClient | None = None,  #: :meta private:
        ignore_tables: list[str] | None = None,
        include_tables: list[str] | None = None,
        sample_rows_in_table_info: int = 3,
        max_string_length: int = 300,
    ) -> None:
        """WatsonxSQLDatabase class."""
        if include_tables and ignore_tables:
            error_msg = "Cannot specify both include_tables and ignore_tables"
            raise ValueError(error_msg)

        self.schema = schema
        self._ignore_tables = set(ignore_tables) if ignore_tables else set()
        self._include_tables = set(include_tables) if include_tables else set()
        self._sample_rows_in_table_info = sample_rows_in_table_info

        self._max_string_length = max_string_length

        if watsonx_client is None:
            url = url or _from_env("WATSONX_URL")
            _validate_param(url, "url", "WATSONX_URL")

            parsed_url = urllib.parse.urlparse(url)
            if parsed_url.netloc.endswith(".cloud.ibm.com"):  # type: ignore[arg-type]
                token = token or _from_env("WATSONX_TOKEN")
                apikey = apikey or _from_env("WATSONX_APIKEY")
                api_key = api_key or _from_env("WATSONX_API_KEY")
                key_to_use = api_key or apikey
                if not token and not key_to_use:
                    error_msg = (
                        "Did not find 'api_key' or 'token',"
                        " please add an environment variable"
                        " `WATSONX_API_KEY` or 'WATSONX_TOKEN' "
                        "which contains it,"
                        " or pass 'api_key' or 'token'"
                        " as a named parameter."
                    )
                    raise ValueError(error_msg)
            else:
                token = token or _from_env("WATSONX_TOKEN")
                apikey = apikey or _from_env("WATSONX_APIKEY")
                api_key = api_key or _from_env("WATSONX_API_KEY")
                key_to_use = api_key or apikey
                password = password or _from_env("WATSONX_PASSWORD")
                if not token and not password and not key_to_use:
                    error_msg = (
                        "Did not find 'token', 'password' or 'api_key',"
                        " please add an environment variable"
                        " `WATSONX_TOKEN`, 'WATSONX_PASSWORD' or 'WATSONX_API_KEY' "
                        "which contains it,"
                        " or pass 'token', 'password' or 'api_key'"
                        " as a named parameter."
                    )
                    raise ValueError(error_msg)

                with contextlib.suppress(ValueError):
                    _validate_param(token, "token", "WATSONX_TOKEN")

                def _check_with_username(
                    auth_name: str,
                    auth_object: Any,
                    username: str | None = username,
                ) -> None:
                    try:
                        _validate_param(
                            auth_object,
                            auth_name,
                            "WATSONX_" + auth_name.upper(),
                        )
                    except ValueError:
                        pass
                    else:
                        username = username or _from_env("WATSONX_USERNAME")
                        _validate_param(username, "username", "WATSONX_USERNAME")

                # validate if password is passed then username is also passed
                _check_with_username(auth_name="password", auth_object=password)

                # validate if api_key is passed then username is also passed
                _check_with_username(auth_name="api_key", auth_object=key_to_use)

                instance_id = instance_id or _from_env("WATSONX_INSTANCE_ID")
                _validate_param(instance_id, "instance_id", "WATSONX_INSTANCE_ID")

            credentials = Credentials(
                url=url,
                api_key=key_to_use,
                token=token,
                password=password,
                username=username,
                instance_id=instance_id,
                version=version,
                verify=verify,
            )
            self.watsonx_client = APIClient(
                credentials=credentials,
                project_id=project_id,
                space_id=space_id,
            )
            project_id = project_id or _from_env("WATSONX_PROJECT_ID")
            space_id = space_id or _from_env("WATSONX_SPACE_ID")
            if project_id:
                self.watsonx_client.set.default_project(project_id=project_id)
            elif space_id:
                self.watsonx_client.set.default_space(space_id=space_id)

        else:
            self.watsonx_client = watsonx_client

        context_id: dict[str, str | None] = {"project_id": None, "space_id": None}
        if project_id is not None:
            context_id["project_id"] = project_id
        elif space_id is not None:
            context_id["space_id"] = space_id
        elif self.watsonx_client.default_project_id is not None:
            context_id["project_id"] = self.watsonx_client.default_project_id
        elif self.watsonx_client.default_space_id is not None:
            context_id["space_id"] = self.watsonx_client.default_space_id
        else:
            error_msg = "Either project_id or space_id is required."
            raise ValueError(error_msg)

        self._flight_sql_client = FlightSQLClient(
            connection_id=connection_id,
            api_client=self.watsonx_client,
            **context_id,
        )

        with self._flight_sql_client as flight_sql_client:
            _tables = flight_sql_client.get_tables(schema=self.schema).get("assets")
            if _tables is not None:
                self._all_tables = {
                    table.get("name") for table in _tables if table.get("name")
                }
            else:
                error_msg = f"No tables found in the schema: {schema}"
                raise RuntimeError(error_msg)

            if self._include_tables:
                missing_tables = self._include_tables - self._all_tables
                if missing_tables:
                    error_msg = f"include_tables {missing_tables} not found in database"
                    raise ValueError(error_msg)
            if self._ignore_tables:
                missing_tables = self._ignore_tables - self._all_tables
                if missing_tables:
                    error_msg = f"ignore_tables {missing_tables} not found in database"
                    raise ValueError(error_msg)

            self._meta_all_tables = {
                table_name: flight_sql_client.get_table_info(
                    table_name=table_name,
                    schema=self.schema,
                    extended_metadata=True,
                    interaction_properties=True,
                )
                for table_name in self._all_tables
                if table_name in (self._include_tables or self._all_tables)
                and table_name not in (self._ignore_tables or {})
            }

    def get_usable_table_names(self) -> Iterable[str]:
        """Get names of tables available."""
        if self._include_tables:
            return sorted(self._include_tables)
        return sorted(self._all_tables - self._ignore_tables)

    def _execute(
        self,
        command: str,
    ) -> Any:
        """Execute a command."""
        with self._flight_sql_client as flight_sql_client:
            results = flight_sql_client.execute(query=command)

        return results.to_dict("records")

    def run(self, command: str, *, include_columns: bool = False) -> str:
        """Execute a SQL command and return a string representing the results."""
        result = self._execute(command)

        res: list[dict[str, Any]] = [
            {
                column: truncate_word(value, length=self._max_string_length)
                for column, value in r.items()
            }
            for r in result
        ]

        if not include_columns:
            res = [tuple(row.values()) for row in res]  # type: ignore[misc]

        return str(res) if res else ""

    def run_no_throw(
        self,
        command: str,
        *,
        include_columns: bool = False,
    ) -> str:
        """Execute a SQL command and return a string representing the results.

        If the statement throws an error, the error message is returned.
        """
        try:
            return self.run(
                command,
                include_columns=include_columns,
            )
        except flight.FlightError as e:
            """Format the error message"""
            return f"Error: {e}"

    def get_table_info(self, table_names: Iterable[str] | None = None) -> str:
        """Get information about specified tables."""
        all_table_names = self.get_usable_table_names()
        if table_names is not None:
            missing_tables = set(table_names).difference(all_table_names)
            if missing_tables:
                error_msg = f"table_names {missing_tables} not found in database"
                raise ValueError(error_msg)

        with self._flight_sql_client as flight_sql_client:
            if table_names is None:
                table_names = self._all_tables

            return "\n\n".join(
                [
                    pretty_print_table_info(
                        schema=self.schema,
                        table_name=table_name,
                        table_info=self._meta_all_tables[table_name],
                    )
                    + f"\n\nFirst {self._sample_rows_in_table_info} rows "
                    + f"of table {table_name}:\n\n"
                    + flight_sql_client.get_n_first_rows(
                        schema=self.schema,
                        table_name=table_name,
                        n=self._sample_rows_in_table_info,
                    ).to_string(index=False)
                    for table_name in table_names
                ],
            )

    def get_table_info_no_throw(self, table_names: Iterable[str] | None = None) -> str:
        """Get information about specified tables."""
        try:
            return self.get_table_info(table_names=table_names)
        except (flight.FlightError, ValueError) as e:
            """Format the error message"""
            return f"Error: {e}"

    def get_context(self) -> dict[str, Any]:
        """Return db context that you may want in agent prompt."""
        table_names = list(self.get_usable_table_names())
        table_info = self.get_table_info_no_throw()
        return {"table_info": table_info, "table_names": ", ".join(table_names)}



"""Test WatsonxLLM API wrapper."""

import os
import re
from unittest.mock import Mock

import pytest
from ibm_watsonx_ai import APIClient  # type: ignore[import-untyped]
from ibm_watsonx_ai.foundation_models.embeddings import (  # type: ignore[import-untyped]
    Embeddings,
)
from ibm_watsonx_ai.gateway import Gateway  # type: ignore[import-untyped]
from ibm_watsonx_ai.wml_client_error import (  # type: ignore[import-untyped]
    WMLClientError,
)

from langchain_ibm import WatsonxEmbeddings

os.environ.pop("WATSONX_APIKEY", None)
os.environ.pop("WATSONX_PROJECT_ID", None)

MODEL_ID = "ibm/granite-embedding-107m-multilingual"

api_client_mock = Mock(spec=APIClient)
api_client_mock.default_space_id = None
api_client_mock.default_project_id = "fake_project_id"

gateway_mock = Mock(spec=Gateway)
gateway_mock._client = api_client_mock

embeddings_mock = Mock(spec=Embeddings)
embeddings_mock._client = api_client_mock
embeddings_mock.model_id = "fake_model_id"
embeddings_mock.params = {"temperature": 1}


def test_initialize_watsonx_embeddings_without_url() -> None:
    pattern = r"(?=.*url)(?=.*WATSONX_URL)"
    with pytest.raises(ValueError, match=pattern):
        WatsonxEmbeddings(
            model_id=MODEL_ID,
        )


def test_initialize_watsonx_embeddings_cloud_only_url() -> None:
    pattern = (
        r"(?=.*api_key)(?=.*token)"
        r"(?=.*WATSONX_API_KEY)(?=.*WATSONX_TOKEN)"
    )
    with pytest.raises(ValueError, match=pattern):
        WatsonxEmbeddings(model_id=MODEL_ID, url="https://us-south.ml.cloud.ibm.com")


def test_initialize_watsonx_embeddings_with_deprecated_apikey() -> None:
    with (
        pytest.warns(
            DeprecationWarning,
            match="'apikey' parameter is deprecated; use 'api_key' instead.",
        ),
        pytest.raises(WMLClientError),
    ):
        WatsonxEmbeddings(
            model_id=MODEL_ID,
            url="https://us-south.ml.cloud.ibm.com",
            apikey="test_apikey",
        )


def test_initialize_watsonx_embeddings_with_api_key_and_apikey() -> None:
    with (
        pytest.warns(
            UserWarning,
            match="Both 'api_key' and deprecated 'apikey' were provided; "
            "'api_key' takes precedence.",
        ),
        pytest.raises(WMLClientError),
    ):
        WatsonxEmbeddings(
            model_id=MODEL_ID,
            url="https://us-south.ml.cloud.ibm.com",
            apikey="fake_apikey",
            api_key="fake_api_key",
        )


def test_initialize_watsonx_embeddings_cpd_without_all() -> None:
    pattern = (
        r"(?=.*api_key)(?=.*password)(?=.*token)"
        r"(?=.*WATSONX_API_KEY)(?=.*WATSONX_PASSWORD)(?=.*WATSONX_TOKEN)"
    )
    with pytest.raises(ValueError, match=pattern):
        WatsonxEmbeddings(
            model_id=MODEL_ID,
            url="https://cpd-zen.apps.cpd48.cp.fyre.ibm.com",
        )


def test_initialize_watsonx_embeddings_cpd_only_password() -> None:
    pattern = r"(?=.*username)(?=.*WATSONX_USERNAME)"
    with pytest.raises(ValueError, match=pattern):
        WatsonxEmbeddings(
            model_id=MODEL_ID,
            url="https://cpd-zen.apps.cpd48.cp.fyre.ibm.com",
            password="test_password",  # noqa: S106
        )


def test_initialize_watsonx_embeddings_cpd_only_apikey() -> None:
    pattern = r"(?=.*username)(?=.*WATSONX_USERNAME)"
    with pytest.raises(ValueError, match=pattern):
        WatsonxEmbeddings(
            model_id=MODEL_ID,
            url="https://cpd-zen.apps.cpd48.cp.fyre.ibm.com",
            apikey="test_apikey",
        )


def test_initialize_watsonx_embeddings_cpd_deprecation_warning_with_instance_id() -> (
    None
):
    with (
        pytest.warns(
            DeprecationWarning, match="The `instance_id` parameter is deprecated"
        ),
        pytest.raises(WMLClientError),
    ):
        WatsonxEmbeddings(
            model_id="google/flan-ul2",
            url="https://cpd-zen.apps.cpd48.cp.fyre.ibm.com",
            apikey="test_apikey",
            username="test_user",
            instance_id="openshift",
        )


def test_initialize_watsonx_embeddings_with_two_exclusive_parameters() -> None:
    pattern = re.escape(
        "The parameters 'model' and 'model_id' are mutually exclusive. "
        "Please specify exactly one of these parameters when initializing "
        "WatsonxEmbeddings."
    )
    with pytest.raises(ValueError, match=pattern):
        WatsonxEmbeddings(
            model_id=MODEL_ID,
            model=MODEL_ID,
            url="https://us-south.ml.cloud.ibm.com",
            apikey="test_apikey",
        )


def test_initialize_watsonx_embeddings_without_any_params() -> None:
    pattern = re.escape(
        "The parameters 'model' and 'model_id' are mutually exclusive. "
        "Please specify exactly one of these parameters when initializing "
        "WatsonxEmbeddings."
    )
    with pytest.raises(ValueError, match=pattern):
        WatsonxEmbeddings()


def test_initialize_watsonx_embeddings_with_api_client_only() -> None:
    pattern = re.escape(
        "The parameters 'model' and 'model_id' are mutually exclusive. "
        "Please specify exactly one of these parameters when initializing "
        "WatsonxEmbeddings."
    )
    with pytest.raises(ValueError, match=pattern):
        WatsonxEmbeddings(watsonx_client=api_client_mock)


def test_initialize_watsonx_embeddings_with_watsonx_embed_gateway() -> None:
    pattern = re.escape(
        "Passing the 'watsonx_embed_gateway' parameter to the WatsonxEmbeddings "
        "constructor is not supported yet."
    )
    with pytest.raises(NotImplementedError, match=pattern):
        WatsonxEmbeddings(watsonx_embed_gateway=gateway_mock)


def test_initialize_watsonx_embeddings_with_watsonx_embed_only() -> None:
    embeddings = WatsonxEmbeddings(watsonx_embed=embeddings_mock)

    assert isinstance(embeddings, WatsonxEmbeddings)



import os
from typing import Any

import pytest
from ibm_watsonx_ai import APIClient  # type: ignore[import-untyped]
from ibm_watsonx_ai.foundation_models.schema import (  # type: ignore[import-untyped]
    RerankParameters,
    RerankReturnOptions,
)
from langchain_core.documents import Document

from langchain_ibm import WatsonxRerank

WX_APIKEY = os.environ.get("WATSONX_APIKEY", "")
WX_PROJECT_ID = os.environ.get("WATSONX_PROJECT_ID", "")

URL = "https://us-south.ml.cloud.ibm.com"

MODEL_ID = "ibm/slate-125m-english-rtrvr-v2"

CREATE_WATSONX_RERANK_INIT_PARAMETERS = [
    pytest.param(
        {
            "model_id": MODEL_ID,
            "url": URL,
            "api_key": WX_APIKEY,
            "project_id": WX_PROJECT_ID,
        },
        id="only api_key",
    ),
    pytest.param(
        {
            "model_id": MODEL_ID,
            "url": URL,
            "apikey": WX_APIKEY,
            "project_id": WX_PROJECT_ID,
        },
        id="only apikey",
    ),
    pytest.param(
        {
            "model_id": MODEL_ID,
            "url": URL,
            "api_key": WX_APIKEY,
            "apikey": WX_APIKEY,
            "project_id": WX_PROJECT_ID,
        },
        id="api_key and apikey",
    ),
]


@pytest.mark.parametrize("init_data", CREATE_WATSONX_RERANK_INIT_PARAMETERS)
def test_01_watsonx_rerank_init(init_data: dict[str, Any]) -> None:
    wx_rerank = WatsonxRerank(**init_data)

    assert isinstance(wx_rerank, WatsonxRerank)


def test_01_rerank_init_with_client() -> None:
    wx_client = APIClient(
        credentials={
            "url": URL,
            "apikey": WX_APIKEY,
        }
    )
    wx_rerank = WatsonxRerank(
        model_id=MODEL_ID, project_id=WX_PROJECT_ID, watsonx_client=wx_client
    )
    assert isinstance(wx_rerank, WatsonxRerank)


def test_02_rerank_documents() -> None:
    wx_rerank = WatsonxRerank(model_id=MODEL_ID, url=URL, project_id=WX_PROJECT_ID)
    test_documents = [
        Document(page_content="This is a test document."),
        Document(page_content="Another test document."),
    ]
    test_query = "Test query"
    results = wx_rerank.rerank(test_documents, test_query)
    assert len(results) == 2


def test_02_rerank_documents_with_params() -> None:
    params = RerankParameters(truncate_input_tokens=2)
    test_documents = [
        Document(page_content="This is a test document."),
    ]
    test_query = "Test query"

    wx_rerank = WatsonxRerank(
        model_id=MODEL_ID,
        url=URL,
        project_id=WX_PROJECT_ID,
    )

    results_1 = wx_rerank.rerank(test_documents, test_query)
    assert len(results_1) == 1

    results_2 = wx_rerank.rerank(test_documents, test_query, params=params)
    assert len(results_2) == 1

    assert results_1[0]["relevance_score"] != results_2[0]["relevance_score"]


def test_03_compress_documents() -> None:
    wx_rerank = WatsonxRerank(
        model_id=MODEL_ID,
        url=URL,
        project_id=WX_PROJECT_ID,
    )
    doc_list = [
        "The Mediterranean diet emphasizes fish, olive oil, and vegetables"
        ", believed to reduce chronic diseases.",
        "Photosynthesis in plants converts light energy into glucose and "
        "produces essential oxygen.",
        "20th-century innovations, from radios to smartphones, centered "
        "on electronic advancements.",
        "Rivers provide water, irrigation, and habitat for aquatic species, "
        "vital for ecosystems.",
        "Apple's conference call to discuss fourth fiscal quarter results and "
        "business updates is scheduled for Thursday, November 2, 2023 at 2:00 "
        "p.m. PT / 5:00 p.m. ET.",
        "Shakespeare's works, like 'Hamlet' and 'A Midsummer Night's Dream,' "
        "endure in literature.",
    ]
    documents = [Document(page_content=x) for x in doc_list]

    result = wx_rerank.compress_documents(
        query="When is the Apple's conference call scheduled?", documents=documents
    )
    assert len(doc_list) == len(result)


def test_04_compress_documents_with_param() -> None:
    params_1 = RerankParameters(
        truncate_input_tokens=1, return_options=RerankReturnOptions(inputs=True)
    )
    params_2 = RerankParameters(truncate_input_tokens=2)
    wx_rerank = WatsonxRerank(
        model_id=MODEL_ID,
        url=URL,
        project_id=WX_PROJECT_ID,
        params=params_1,
    )
    doc_list = [
        "The Mediterranean diet emphasizes fish, olive oil, and vegetables"
        ", believed to reduce chronic diseases.",
        "Photosynthesis in plants converts light energy into glucose and "
        "produces essential oxygen.",
        "20th-century innovations, from radios to smartphones, centered "
        "on electronic advancements.",
        "Rivers provide water, irrigation, and habitat for aquatic species, "
        "vital for ecosystems.",
        "Apple's conference call to discuss fourth fiscal quarter results and "
        "business updates is scheduled for Thursday, November 2, 2023 at 2:00 "
        "p.m. PT / 5:00 p.m. ET.",
        "Shakespeare's works, like 'Hamlet' and 'A Midsummer Night's Dream,' "
        "endure in literature.",
    ]
    documents = [Document(page_content=x) for x in doc_list]

    result_1 = wx_rerank.compress_documents(
        query="When is the Apple's conference call scheduled?", documents=documents
    )
    assert len(doc_list) == len(result_1)
    result_2 = wx_rerank.compress_documents(
        query="When is the Apple's conference call scheduled?",
        documents=documents,
        params=params_2,
    )
    assert len(doc_list) == len(result_2)

    assert next(
        el.metadata["relevance_score"]
        for el in result_1
        if el.page_content.startswith("20th")
    ) != next(
        el.metadata["relevance_score"]
        for el in result_2
        if el.page_content.startswith("20th")
    )



import os
from typing import Any, Literal

import pytest
from langchain_core.language_models import BaseChatModel
from langchain_tests.integration_tests.chat_models import ChatModelIntegrationTests
from pytest_benchmark.fixture import BenchmarkFixture
from vcr.cassette import Cassette  # type: ignore[import-untyped]

from langchain_ibm import ChatWatsonx

WX_APIKEY = os.environ.get("WATSONX_APIKEY", "")
WX_PROJECT_ID = os.environ.get("WATSONX_PROJECT_ID", "")

URL = "https://us-south.ml.cloud.ibm.com"

MODEL_ID = "ibm/granite-3-3-8b-instruct"
MODEL_ID_IMAGE = "meta-llama/llama-3-2-90b-vision-instruct"


class TestChatWatsonxStandard(ChatModelIntegrationTests):
    @property
    def chat_model_class(self) -> type[BaseChatModel]:
        return ChatWatsonx

    @property
    def supports_image_inputs(self) -> bool:
        return True

    @property
    def supports_json_mode(self) -> bool:
        return True

    @property
    def enable_vcr_tests(self) -> bool:
        return False

    @property
    def supported_usage_metadata_details(
        self,
    ) -> dict[
        Literal["invoke", "stream"],
        list[
            Literal[
                "audio_input",
                "audio_output",
                "reasoning_output",
                "cache_read_input",
                "cache_creation_input",
            ]
        ],
    ]:
        return {
            "invoke": [],
            "stream": [],
        }

    @property
    def chat_model_params(self) -> dict[str, Any]:
        return {
            "model_id": MODEL_ID,
            "url": URL,
            "apikey": WX_APIKEY,
            "project_id": WX_PROJECT_ID,
        }

    @property
    def standard_chat_model_params(self) -> dict[str, Any]:
        return {
            "temperature": 0,
            "params": {
                "include_reasoning": False,
            },
        }

    @pytest.mark.xfail(reason="Supported for vision model.")
    def test_image_inputs(self, model: BaseChatModel) -> None:
        model.watsonx_model._inference.model_id = MODEL_ID_IMAGE  # type: ignore[attr-defined]
        super().test_image_inputs(model)
        model.watsonx_model._inference.model_id = MODEL_ID  # type: ignore[attr-defined]

    @pytest.mark.xfail(reason="Unsupported for v1.")
    @pytest.mark.parametrize("model", [{}, {"output_version": "v1"}], indirect=True)
    def test_agent_loop(self, model: BaseChatModel) -> None:
        super().test_agent_loop(model)

    @pytest.mark.xfail(reason="VCR not set up.")
    def test_stream_time(
        self, model: BaseChatModel, benchmark: BenchmarkFixture, vcr: Cassette
    ) -> None:
        super().test_stream_time(model, benchmark, vcr)



"""Test WatsonxToolkit.

You'll need to set WATSONX_APIKEY environment variable.
"""

import json
import os
from typing import Any

import pytest

from langchain_ibm.agent_toolkits.utility import WatsonxToolkit

WX_APIKEY = os.environ.get("WATSONX_APIKEY", "")

URL = "https://us-south.ml.cloud.ibm.com"

TOOL_NAME_1 = "GoogleSearch"
TOOL_NAME_2 = "Weather"

CREATE_CHAT_WATSONX_TOOLKIT_PARAMETERS = [
    pytest.param(
        {
            "url": URL,
            "api_key": WX_APIKEY,
        },
        id="only api_key",
    ),
    pytest.param(
        {
            "url": URL,
            "apikey": WX_APIKEY,
        },
        id="only apikey",
    ),
    pytest.param(
        {
            "url": URL,
            "api_key": WX_APIKEY,
            "apikey": WX_APIKEY,
        },
        id="api_key and apikey",
    ),
]


@pytest.mark.parametrize("init_data", CREATE_CHAT_WATSONX_TOOLKIT_PARAMETERS)
def test_00_watsonx_toolkit_init(init_data: dict[str, Any]) -> None:
    watsonx_toolkit = WatsonxToolkit(**init_data)

    assert isinstance(watsonx_toolkit, WatsonxToolkit)


def test_01_get_tools() -> None:
    watsonx_toolkit = WatsonxToolkit(
        url=URL,
    )
    tools = watsonx_toolkit.get_tools()
    assert tools


def test_02_get_tool_with_config_schema() -> None:
    watsonx_toolkit = WatsonxToolkit(
        url=URL,
    )
    tool = watsonx_toolkit.get_tool(tool_name=TOOL_NAME_1)
    assert tool.name == TOOL_NAME_1
    assert tool.description
    assert tool.tool_config_schema


def test_03_get_tool_with_input_schema() -> None:
    watsonx_toolkit = WatsonxToolkit(
        url=URL,
    )
    tool = watsonx_toolkit.get_tool(tool_name=TOOL_NAME_2)
    assert tool.name == TOOL_NAME_2
    assert tool.description
    assert tool.tool_input_schema


def test_04_invoke_tool_with_config_schema() -> None:
    watsonx_toolkit = WatsonxToolkit(
        url=URL,
    )
    tool = watsonx_toolkit.get_tool(tool_name=TOOL_NAME_1)

    config = {
        "maxResults": 3,
    }
    tool.set_tool_config(config)

    tool_input = {
        "q": "Search IBM",
    }

    result = tool.invoke(tool_input)
    output = json.loads(result.get("output"))

    assert isinstance(output, list)
    assert len(output) == 3
    for answer in output:
        assert answer.get("title")
        assert answer.get("description")
        assert answer.get("url")


def test_05_invoke_tool_with_input_schema() -> None:
    watsonx_toolkit = WatsonxToolkit(
        url=URL,
    )
    tool = watsonx_toolkit.get_tool(tool_name=TOOL_NAME_2)

    tool_input = {
        "location": "New York",
    }

    result = tool.invoke(tool_input)
    output = result.get("output")

    assert output
    assert tool_input["location"] in output
    assert "temperature" in output.lower()


def test_06_invoke_tool_with_simple_input() -> None:
    watsonx_toolkit = WatsonxToolkit(
        url=URL,
    )
    tool = watsonx_toolkit.get_tool(tool_name=TOOL_NAME_1)

    tool_input = {
        "q": "Search IBM",
    }

    result = tool.invoke(tool_input)
    output = json.loads(result.get("output"))

    assert isinstance(output, list)
    for answer in output:
        assert answer.get("title")
        assert answer.get("description")
        assert answer.get("url")
